"""Defines Blink cameras."""

import copy
import string
import os
import logging
import datetime
from json import dumps
import traceback
import aiohttp
from aiofiles import open
from requests.compat import urljoin
from blinkpy import api
from blinkpy.helpers.constants import TIMEOUT_MEDIA
from blinkpy.helpers.util import to_alphanumeric

_LOGGER = logging.getLogger(__name__)


class BlinkCamera:
    """Class to initialize individual camera."""

    def __init__(self, sync):
        """Initialize BlinkCamera."""
        self.sync = sync
        self.name = None
        self.camera_id = None
        self.network_id = None
        self.thumbnail = None
        self.serial = None
        self._version = None
        self.motion_enabled = None
        self.battery_level = None
        self._battery_voltage = None
        self.clip = None
        # A clip remains in the recent clips list until is has
        # been downloaded or has been expired.
        self.recent_clips = []
        self.temperature = None
        self.temperature_calibrated = None
        self.battery_state = None
        self.motion_detected = None
        self.wifi_strength = None
        self.last_record = None
        self._cached_image = None
        self._cached_video = None
        self.camera_type = ""
        self.product_type = None
        self.sync_signal_strength = None

    @property
    def attributes(self):
        """Return dictionary of all camera attributes."""
        attributes = {
            "name": self.name,
            "camera_id": self.camera_id,
            "serial": self.serial,
            "version": self._version,
            "temperature": self.temperature,
            "temperature_c": self.temperature_c,
            "temperature_calibrated": self.temperature_calibrated,
            "battery": self.battery,
            "battery_level": self.battery_level,
            "battery_voltage": self._battery_voltage,
            "thumbnail": self.thumbnail,
            "video": self.clip,
            "recent_clips": self.recent_clips,
            "motion_enabled": self.motion_enabled,
            "motion_detected": self.motion_detected,
            "wifi_strength": self.wifi_strength,
            "network_id": self.sync.network_id,
            "sync_module": self.sync.name,
            "sync_signal_strength": self.sync_signal_strength,
            "last_record": self.last_record,
            "type": self.product_type,
        }
        return attributes

    @property
    def battery(self):
        """Return battery as string."""
        return self.battery_state

    @property
    def battery_voltage(self):
        """Return battery voltage as a number in 100ths of volts, so 165 = 1.65v."""
        return self._battery_voltage

    @property
    def temperature_c(self):
        """Return temperature in celsius."""
        try:
            return round((self.temperature - 32) / 9.0 * 5.0, 1)
        except TypeError:
            return None

    @property
    def image_from_cache(self):
        """Return the most recently cached image."""
        if self._cached_image:
            return self._cached_image
        return None

    @property
    def video_from_cache(self):
        """Return the most recently cached video."""
        if self._cached_video:
            return self._cached_video
        return None

    @property
    def version(self):
        """Return the camera Firmware version."""
        return self._version

    @property
    def arm(self):
        """Return arm status of camera."""
        return self.motion_enabled

    async def async_arm(self, value):
        """Set camera arm status."""
        if value:
            return await api.request_motion_detection_enable(
                self.sync.blink, self.network_id, self.camera_id
            )
        return await api.request_motion_detection_disable(
            self.sync.blink, self.network_id, self.camera_id
        )

    @property
    async def night_vision(self):
        """Return night_vision status."""
        res = await api.request_get_config(
            self.sync.blink,
            self.network_id,
            self.camera_id,
            product_type=self.product_type,
        )
        if res is None:
            return None
        if self.product_type == "catalina":
            res = res.get("camera", [{}])[0]
        if res["illuminator_enable"] in [0, 1, 2]:
            res["illuminator_enable"] = ["off", "on", "auto"][
                res.get("illuminator_enable")
            ]
        nv_keys = [
            "night_vision_control",
            "illuminator_enable",
            "illuminator_enable_v2",
        ]
        return {key: res.get(key) for key in nv_keys}

    async def async_set_night_vision(self, value):
        """Set camera night_vision status."""
        if value not in ["on", "off", "auto"]:
            return None
        if self.product_type == "catalina":
            value = {"off": 0, "on": 1, "auto": 2}.get(value, None)
        data = dumps({"illuminator_enable": value})
        res = await api.request_update_config(
            self.sync.blink,
            self.network_id,
            self.camera_id,
            product_type=self.product_type,
            data=data,
        )
        if res and res.status == 200:
            return await res.json()
        return None

    async def record(self):
        """Initiate clip recording."""
        return await api.request_new_video(
            self.sync.blink, self.network_id, self.camera_id
        )

    async def get_media(self, media_type="image") -> aiohttp.ClientRequest:
        """Download media (image or video)."""
        if media_type.lower() == "video":
            return await self.get_video_clip()
        return await self.get_thumbnail()

    async def get_thumbnail(self, url=None):
        """Download thumbnail image."""
        if not url:
            url = self.thumbnail
            if not url:
                _LOGGER.warning("Thumbnail URL not available: self.thumbnail=%s", url)
                return None
        return await api.http_get(
            self.sync.blink,
            url=url,
            stream=True,
            json=False,
            timeout=TIMEOUT_MEDIA,
        )

    async def get_video_clip(self, url=None):
        """Download video clip."""
        if not url:
            url = self.clip
            if not url:
                _LOGGER.warning("Video clip URL not available: self.clip=%s", url)
                return None
        return await api.http_get(
            self.sync.blink,
            url=url,
            stream=True,
            json=False,
            timeout=TIMEOUT_MEDIA,
        )

    async def snap_picture(self):
        """Take a picture with camera to create a new thumbnail."""
        ret_val = await api.request_new_image(
            self.sync.blink, self.network_id, self.camera_id
        )
        response = await self.get_media()
        if response and response.status == 200:
            self._cached_image = await response.read()

        return ret_val

    async def set_motion_detect(self, enable):
        """Set motion detection."""
        _LOGGER.warning(
            "Method is deprecated as of v0.16.0 and will be removed in "
            "a future version. Please use the BlinkCamera.arm property instead."
        )
        if enable:
            return await api.request_motion_detection_enable(
                self.sync.blink, self.network_id, self.camera_id
            )
        return await api.request_motion_detection_disable(
            self.sync.blink, self.network_id, self.camera_id
        )

    async def update(self, config, force_cache=False, expire_clips=True, **kwargs):
        """Update camera info."""
        if bool(config):
            self.extract_config_info(config)
            await self.get_sensor_info()
            await self.update_images(
                config, force_cache=force_cache, expire_clips=expire_clips
            )

    def extract_config_info(self, config):
        """Extract info from config."""
        self.name = config.get("name", "unknown")
        self.camera_id = str(config.get("id", "unknown"))
        self.network_id = str(config.get("network_id", "unknown"))
        self.serial = config.get("serial")
        self._version = config.get("fw_version")
        self.motion_enabled = config.get("enabled", "unknown")
        self._battery_voltage = config.get("battery_voltage", None)
        self.battery_state = config.get("battery_state") or config.get("battery")
        self.wifi_strength = config.get("wifi_strength")
        if signals := config.get("signals"):
            self.battery_level = signals.get("battery")
            self.sync_signal_strength = signals.get("lfr")
            self.temperature = signals.get("temp")
        else:
            self.temperature = config.get("temperature")
        self.product_type = config.get("type")

    async def get_sensor_info(self):
        """Retrieve calibrated temperature from special endpoint."""
        resp = await api.request_camera_sensors(
            self.sync.blink, self.network_id, self.camera_id
        )
        try:
            self.temperature_calibrated = resp["temp"]
        except (TypeError, KeyError):
            self.temperature_calibrated = self.temperature
            _LOGGER.warning(
                "Could not retrieve calibrated temperature response %s.", resp
            )
            _LOGGER.warning(
                "for network_id (%s) and camera_id (%s)",
                self.network_id,
                self.camera_id,
            )

    async def update_images(self, config, force_cache=False, expire_clips=True):
        """Update images for camera."""
        new_thumbnail = None
        thumb_addr = None
        thumb_string = None
        if config.get("thumbnail", False):
            thumb_addr = config["thumbnail"]
            try:
                # API update only returns the timestamp!
                int(thumb_addr)
                thumb_string = (
                    "/api/v3/media/accounts/"
                    f"{self.sync.blink.account_id}/networks/"
                    f"{self.network_id}/{self.product_type}/"
                    f"{self.camera_id}/thumbnail/"
                    f"thumbnail.jpg?ts={thumb_addr}&ext="
                )
            except ValueError:
                # This is the old API and has the full url
                thumb_string = f"{thumb_addr}.jpg"
                # Check that new full api url has not been returned:
                if thumb_addr.endswith("&ext="):
                    thumb_string = thumb_addr

            if thumb_string is not None:
                new_thumbnail = urljoin(self.sync.urls.base_url, thumb_string)

        else:
            _LOGGER.warning("Could not find thumbnail for camera %s.", self.name)

        try:
            self.motion_detected = self.sync.motion[self.name]
        except KeyError:
            self.motion_detected = False

        clip_addr = None
        try:

            def timesort(record):
                rec_time = record["time"]
                iso_time = datetime.datetime.fromisoformat(rec_time)
                stamp = int(iso_time.timestamp())
                return stamp

            if (
                len(self.sync.last_records) > 0
                and len(self.sync.last_records[self.name]) > 0
            ):
                last_records = sorted(self.sync.last_records[self.name], key=timesort)
                for rec in last_records:
                    clip_addr = rec["clip"]
                    self.clip = f"{self.sync.urls.base_url}{clip_addr}"
                    self.last_record = rec["time"]
                    if self.motion_detected:
                        recent = {"time": self.last_record, "clip": self.clip}
                        # Prevent duplicates.
                        if recent not in self.recent_clips:
                            self.recent_clips.append(recent)
                if len(self.recent_clips) > 0:
                    _LOGGER.debug(
                        "Found %s recent clips for %s",
                        len(self.recent_clips),
                        self.name,
                    )
                    _LOGGER.debug(
                        "Most recent clip for %s was created at %s : %s",
                        self.name,
                        self.last_record,
                        self.clip,
                    )
        except (KeyError, IndexError):
            ex = traceback.format_exc()
            trace = "".join(traceback.format_stack())
            _LOGGER.error("Error getting last records for '%s': %s", self.name, ex)
            _LOGGER.debug("\n%s", trace)

        # If the thumbnail or clip have changed, update the cache
        update_cached_image = False
        if new_thumbnail != self.thumbnail or self._cached_image is None:
            update_cached_image = True
        self.thumbnail = new_thumbnail

        update_cached_video = False
        if self._cached_video is None or self.motion_detected:
            update_cached_video = True

        if new_thumbnail is not None and (update_cached_image or force_cache):
            response = await self.get_media()
            if response and response.status == 200:
                self._cached_image = await response.read()

        if clip_addr is not None and (update_cached_video or force_cache):
            response = await self.get_media(media_type="video")
            if response and response.status == 200:
                self._cached_video = await response.read()

        # Don't let the recent clips list grow without bound.
        if expire_clips:
            await self.expire_recent_clips()

    async def expire_recent_clips(self, delta=datetime.timedelta(hours=1)):
        """Remove recent clips from list when they get too old."""
        to_keep = []
        for clip in self.recent_clips:
            timedelta = (datetime.datetime.now() - delta).timestamp()
            clip_time = datetime.datetime.fromisoformat(clip["time"]).timestamp()
            if clip_time > timedelta:
                to_keep.append(clip)
        num_expired = len(self.recent_clips) - len(to_keep)
        if num_expired > 0:
            _LOGGER.info("Expired %s clips from '%s'", num_expired, self.name)
        self.recent_clips = copy.deepcopy(to_keep)
        if len(self.recent_clips) > 0:
            _LOGGER.info(
                "'%s' has %s clips available for download",
                self.name,
                len(self.recent_clips),
            )
            for clip in self.recent_clips:
                url = clip["clip"]
                if "local_storage" in url:
                    await api.http_post(self.sync.blink, url)

    async def get_liveview(self):
        """Get liveview rtsps link."""
        response = await api.request_camera_liveview(
            self.sync.blink, self.sync.network_id, self.camera_id
        )
        return response["server"]

    async def image_to_file(self, path):
        """
        Write image to file.

        :param path: Path to write file
        """
        _LOGGER.debug("Writing image from %s to %s", self.name, path)
        response = await self.get_media()
        if response and response.status == 200:
            async with open(path, "wb") as imagefile:
                await imagefile.write(await response.read())
        else:
            _LOGGER.error("Cannot write image to file, response %s", response.status)

    async def video_to_file(self, path):
        """
        Write video to file.

        :param path: Path to write file
        """
        _LOGGER.debug("Writing video from %s to %s", self.name, path)
        response = await self.get_media(media_type="video")
        if response is None:
            _LOGGER.error("No saved video exists for %s.", self.name)
            return
        async with open(path, "wb") as vidfile:
            await vidfile.write(await response.read())

    async def save_recent_clips(
        self, output_dir="/tmp", file_pattern="${created}_${name}.mp4"
    ):
        """Save all recent clips using timestamp file name pattern."""
        if output_dir[-1] != "/":
            output_dir += "/"

        recent = copy.deepcopy(self.recent_clips)

        num_saved = 0
        for clip in recent:
            clip_time = datetime.datetime.fromisoformat(clip["time"])
            clip_time_local = clip_time.replace(
                tzinfo=datetime.timezone.utc
            ).astimezone(tz=None)
            created_at = clip_time_local.strftime("%Y%m%d_%H%M%S")
            clip_addr = clip["clip"]

            file_name = string.Template(file_pattern).substitute(
                created=created_at, name=to_alphanumeric(self.name)
            )
            path = os.path.join(output_dir, file_name)
            _LOGGER.debug("Saving %s to %s", clip_addr, path)
            media = await self.get_video_clip(clip_addr)
            if media and media.status == 200:
                async with open(path, "wb") as clip_file:
                    await clip_file.write(await media.read())
                num_saved += 1
                try:
                    # Remove recent clip from the list once the download has finished.
                    self.recent_clips.remove(clip)
                    _LOGGER.debug("Removed %s from recent clips", clip)
                except ValueError:
                    ex = traceback.format_exc()
                    _LOGGER.error("Error removing clip from list: %s", ex)
                    trace = "".join(traceback.format_stack())
                    _LOGGER.debug("\n%s", trace)

        if len(recent) == 0:
            _LOGGER.info("No recent clips to save for '%s'.", self.name)
        else:
            _LOGGER.info(
                "Saved %s of %s recent clips from '%s' to directory %s",
                num_saved,
                len(recent),
                self.name,
                output_dir,
            )


class BlinkCameraMini(BlinkCamera):
    """Define a class for a Blink Mini camera."""

    def __init__(self, sync):
        """Initialize a Blink Mini cameras."""
        super().__init__(sync)
        self.camera_type = "mini"

    @property
    def arm(self):
        """Return camera arm status."""
        return self.sync.arm

    async def async_arm(self, value):
        """Set camera arm status."""
        url = (
            f"{self.sync.urls.base_url}/api/v1/accounts/"
            f"{self.sync.blink.account_id}/networks/"
            f"{self.network_id}/owls/{self.camera_id}/config"
        )
        data = dumps({"enabled": value})
        response = await api.http_post(self.sync.blink, url, data=data)
        await api.wait_for_command(self.sync.blink, response)
        return response

    async def record(self):
        """Initiate clip recording for a blink mini camera."""
        url = (
            f"{self.sync.urls.base_url}/api/v1/accounts/"
            f"{self.sync.blink.account_id}/networks/"
            f"{self.network_id}/owls/{self.camera_id}/clip"
        )
        response = await api.http_post(self.sync.blink, url)
        await api.wait_for_command(self.sync.blink, response)
        return response

    async def snap_picture(self):
        """Snap picture for a blink mini camera."""
        url = (
            f"{self.sync.urls.base_url}/api/v1/accounts/"
            f"{self.sync.blink.account_id}/networks/"
            f"{self.network_id}/owls/{self.camera_id}/thumbnail"
        )
        response = await api.http_post(self.sync.blink, url)
        await api.wait_for_command(self.sync.blink, response)
        return response

    async def get_sensor_info(self):
        """Get sensor info for blink mini camera."""

    async def get_liveview(self):
        """Get liveview link."""
        url = (
            f"{self.sync.urls.base_url}/api/v1/accounts/"
            f"{self.sync.blink.account_id}/networks/"
            f"{self.network_id}/owls/{self.camera_id}/liveview"
        )
        response = await api.http_post(self.sync.blink, url)
        await api.wait_for_command(self.sync.blink, response)
        server = response["server"]
        server_split = server.split(":")
        server_split[0] = "rtsps"
        link = ":".join(server_split)
        return link


class BlinkDoorbell(BlinkCamera):
    """Define a class for a Blink Doorbell camera."""

    def __init__(self, sync):
        """Initialize a Blink Doorbell."""
        super().__init__(sync)
        self.camera_type = "doorbell"

    @property
    def arm(self):
        """Return camera arm status."""
        return self.motion_enabled

    async def async_arm(self, value):
        """Set camera arm status."""
        url = (
            f"{self.sync.urls.base_url}/api/v1/accounts/"
            f"{self.sync.blink.account_id}/networks/"
            f"{self.sync.network_id}/doorbells/{self.camera_id}"
        )
        if value:
            url = f"{url}/enable"
        else:
            url = f"{url}/disable"

        response = await api.http_post(self.sync.blink, url)
        await api.wait_for_command(self.sync.blink, response)
        return response

    async def record(self):
        """Initiate clip recording for a blink doorbell camera."""
        url = (
            f"{self.sync.urls.base_url}/api/v1/accounts/"
            f"{self.sync.blink.account_id}/networks/"
            f"{self.sync.network_id}/doorbells/{self.camera_id}/clip"
        )

        response = await api.http_post(self.sync.blink, url)
        await api.wait_for_command(self.sync.blink, response)
        return response

    async def snap_picture(self):
        """Snap picture for a blink doorbell camera."""
        url = (
            f"{self.sync.urls.base_url}/api/v1/accounts/"
            f"{self.sync.blink.account_id}/networks/"
            f"{self.sync.network_id}/doorbells/{self.camera_id}/thumbnail"
        )

        response = await api.http_post(self.sync.blink, url)
        await api.wait_for_command(self.sync.blink, response)
        return response

    async def get_sensor_info(self):
        """Get sensor info for blink doorbell camera."""

    async def get_liveview(self):
        """Get liveview link."""
        url = (
            f"{self.sync.urls.base_url}/api/v1/accounts/"
            f"{self.sync.blink.account_id}/networks/"
            f"{self.sync.network_id}/doorbells/{self.camera_id}/liveview"
        )
        response = await api.http_post(self.sync.blink, url)
        await api.wait_for_command(self.sync.blink, response)
        server = response["server"]
        link = server.replace("immis://", "rtsps://")
        return link



"""Tests camera and system functions."""

from unittest import mock, IsolatedAsyncioTestCase
import time
import random
from io import BufferedIOBase
import aiofiles
from blinkpy import blinkpy
from blinkpy.sync_module import BlinkSyncModule
from blinkpy.camera import BlinkCamera
from blinkpy.helpers.util import get_time, BlinkURLHandler


class MockSyncModule(BlinkSyncModule):
    """Mock blink sync module object."""

    async def get_network_info(self):
        """Mock network info method."""
        return True


class MockCamera(BlinkCamera):
    """Mock blink camera object."""

    def __init__(self, sync):
        """Initialize mock camera."""
        super().__init__(sync)
        self.camera_id = random.randint(1, 100000)

    async def update(self, config, force_cache=False, **kwargs):
        """Mock camera update method."""


class TestBlinkFunctions(IsolatedAsyncioTestCase):
    """Test Blink and BlinkCamera functions in blinkpy."""

    def setUp(self):
        """Set up Blink module."""
        self.blink = blinkpy.Blink(session=mock.AsyncMock())
        self.blink.urls = BlinkURLHandler("test")

    def tearDown(self):
        """Clean up after test."""
        self.blink = None

    def test_merge_cameras(self):
        """Test merge camera functionality."""
        first_dict = {"foo": "bar", "test": 123}
        next_dict = {"foobar": 456, "bar": "foo"}
        self.blink.sync["foo"] = BlinkSyncModule(self.blink, "foo", 1, [])
        self.blink.sync["bar"] = BlinkSyncModule(self.blink, "bar", 2, [])
        self.blink.sync["foo"].cameras = first_dict
        self.blink.sync["bar"].cameras = next_dict
        result = self.blink.merge_cameras()
        expected = {"foo": "bar", "test": 123, "foobar": 456, "bar": "foo"}
        self.assertEqual(expected, result)

    @mock.patch("blinkpy.blinkpy.api.request_videos")
    async def test_download_video_exit(self, mock_req):
        """Test we exit method when provided bad response."""
        blink = blinkpy.Blink(session=mock.AsyncMock())
        blink.last_refresh = 0
        mock_req.return_value = {}
        formatted_date = get_time(blink.last_refresh)
        expected_log = [
            f"INFO:blinkpy.blinkpy:Retrieving videos since {formatted_date}",
            "DEBUG:blinkpy.blinkpy:Processing page 1",
            "INFO:blinkpy.blinkpy:No videos found on page 1. Exiting.",
        ]
        with self.assertLogs(level="DEBUG") as dl_log:
            await blink.download_videos("/tmp")
        self.assertListEqual(dl_log.output, expected_log)

    @mock.patch("blinkpy.blinkpy.api.request_videos")
    async def test_parse_downloaded_items(self, mock_req):
        """Test ability to parse downloaded items list."""
        blink = blinkpy.Blink(session=mock.AsyncMock())
        generic_entry = {
            "created_at": "1970",
            "device_name": "foo",
            "deleted": True,
            "media": "/bar.mp4",
        }
        result = [generic_entry]
        mock_req.return_value = {"media": result}
        blink.last_refresh = 0
        formatted_date = get_time(blink.last_refresh)
        expected_log = [
            f"INFO:blinkpy.blinkpy:Retrieving videos since {formatted_date}",
            "DEBUG:blinkpy.blinkpy:Processing page 1",
            "DEBUG:blinkpy.blinkpy:foo: /bar.mp4 is marked as deleted.",
        ]
        with self.assertLogs(level="DEBUG") as dl_log:
            await blink.download_videos("/tmp", stop=2, delay=0)
        self.assertListEqual(dl_log.output, expected_log)

    @mock.patch("blinkpy.blinkpy.api.request_videos")
    async def test_parse_downloaded_throttle(self, mock_req):
        """Test ability to parse downloaded items list."""
        generic_entry = {
            "created_at": "1970",
            "device_name": "foo",
            "deleted": False,
            "media": "/bar.mp4",
        }
        result = [generic_entry]
        mock_req.return_value = {"media": result}
        self.blink.last_refresh = 0
        start = time.time()
        await self.blink.download_videos("/tmp", stop=2, delay=0, debug=True)
        now = time.time()
        delta = now - start
        self.assertTrue(delta < 0.1)

        start = time.time()
        await self.blink.download_videos("/tmp", stop=2, delay=0.1, debug=True)
        now = time.time()
        delta = now - start
        self.assertTrue(delta >= 0.1)

    @mock.patch("blinkpy.blinkpy.api.request_videos")
    async def test_get_videos_metadata(self, mock_req):
        """Test ability to fetch videos metadata."""
        generic_entry = {
            "created_at": "1970",
            "device_name": "foo",
            "deleted": True,
            "media": "/bar.mp4",
        }
        result = [generic_entry]
        mock_req.return_value = {"media": result}
        self.blink.last_refresh = 0

        results = await self.blink.get_videos_metadata(stop=2)
        self.assertListEqual(results, result)

        results = await self.blink.get_videos_metadata(
            since="2018/07/28 12:33:00", stop=2
        )
        self.assertListEqual(results, result)

        mock_req.return_value = {"media": None}
        results = await self.blink.get_videos_metadata(stop=2)
        self.assertListEqual(results, [])

    @mock.patch("blinkpy.blinkpy.api.http_get")
    async def test_do_http_get(self, mock_req):
        """Test ability to do_http_get."""
        blink = blinkpy.Blink(session=mock.AsyncMock())
        blink.urls = BlinkURLHandler("test")
        response = await blink.do_http_get("/path/to/request")
        self.assertTrue(response is not None)

    @mock.patch("blinkpy.blinkpy.api.request_videos")
    async def test_download_videos_deleted(self, mock_req):
        """Test ability to download videos."""
        generic_entry = {
            "created_at": "1970",
            "device_name": "foo",
            "deleted": True,
            "media": "/bar.mp4",
        }
        result = [generic_entry]
        mock_req.return_value = {"media": result}
        self.blink.last_refresh = 0
        formatted_date = get_time(self.blink.last_refresh)
        expected_log = [
            f"INFO:blinkpy.blinkpy:Retrieving videos since {formatted_date}",
            "DEBUG:blinkpy.blinkpy:Processing page 1",
            "DEBUG:blinkpy.blinkpy:foo: /bar.mp4 is marked as deleted.",
        ]
        with self.assertLogs(level="DEBUG") as dl_log:
            await self.blink.download_videos("/tmp", camera="foo", stop=2, delay=0)
        self.assertListEqual(dl_log.output, expected_log)

    @mock.patch("blinkpy.blinkpy.api.request_videos")
    @mock.patch("aiofiles.ospath.isfile")
    async def test_download_videos_file(self, mock_isfile, mock_req):
        """Test ability to download videos to a file."""
        generic_entry = {
            "created_at": "1970",
            "device_name": "foo",
            "deleted": False,
            "media": "/bar.mp4",
        }
        result = [generic_entry]
        mock_req.return_value = {"media": result}
        mock_isfile.return_value = False
        self.blink.last_refresh = 0

        aiofiles.threadpool.wrap.register(mock.MagicMock)(
            lambda *args, **kwargs: aiofiles.threadpool.AsyncBufferedIOBase(
                *args, **kwargs
            )
        )
        mock_file = mock.MagicMock(spec=BufferedIOBase)
        with mock.patch("aiofiles.threadpool.sync_open", return_value=mock_file):
            await self.blink.download_videos("/tmp", camera="foo", stop=2, delay=0)
            assert mock_file.write.call_count == 1

    @mock.patch("blinkpy.blinkpy.api.request_videos")
    @mock.patch("aiofiles.ospath.isfile")
    async def test_download_videos_file_exists(self, mock_isfile, mock_req):
        """Test ability to download videos with file exists."""
        generic_entry = {
            "created_at": "1970",
            "device_name": "foo",
            "deleted": False,
            "media": "/bar.mp4",
        }
        result = [generic_entry]
        mock_req.return_value = {"media": result}
        mock_isfile.return_value = True

        self.blink.last_refresh = 0
        formatted_date = get_time(self.blink.last_refresh)
        expected_log = [
            f"INFO:blinkpy.blinkpy:Retrieving videos since {formatted_date}",
            "DEBUG:blinkpy.blinkpy:Processing page 1",
            "INFO:blinkpy.blinkpy:/tmp/foo-1970.mp4 already exists, skipping...",
        ]
        with self.assertLogs(level="DEBUG") as dl_log:
            await self.blink.download_videos("/tmp", camera="foo", stop=2, delay=0)
            assert expected_log[0] in dl_log.output
            assert expected_log[1] in dl_log.output
            assert expected_log[2] in dl_log.output

    @mock.patch("blinkpy.blinkpy.api.request_videos")
    async def test_parse_camera_not_in_list(self, mock_req):
        """Test ability to parse downloaded items list."""
        generic_entry = {
            "created_at": "1970",
            "device_name": "foo",
            "deleted": True,
            "media": "/bar.mp4",
        }
        result = [generic_entry]
        mock_req.return_value = {"media": result}
        self.blink.last_refresh = 0
        formatted_date = get_time(self.blink.last_refresh)
        expected_log = [
            f"INFO:blinkpy.blinkpy:Retrieving videos since {formatted_date}",
            "DEBUG:blinkpy.blinkpy:Processing page 1",
            "DEBUG:blinkpy.blinkpy:Skipping videos for foo.",
        ]
        with self.assertLogs(level="DEBUG") as dl_log:
            await self.blink.download_videos("/tmp", camera="bar", stop=2, delay=0)
        self.assertListEqual(dl_log.output, expected_log)

    @mock.patch("blinkpy.blinkpy.api.request_videos")
    async def test_parse_malformed_entry(self, mock_req):
        """Test ability to parse downloaded items in malformed list."""
        self.blink.last_refresh = 0
        formatted_date = get_time(self.blink.last_refresh)
        generic_entry = {
            "created_at": "1970",
        }
        result = [generic_entry]
        mock_req.return_value = {"media": result}
        expected_log = [
            f"INFO:blinkpy.blinkpy:Retrieving videos since {formatted_date}",
            "DEBUG:blinkpy.blinkpy:Processing page 1",
            "INFO:blinkpy.blinkpy:Missing clip information, skipping...",
        ]
        with self.assertLogs(level="DEBUG") as dl_log:
            await self.blink.download_videos("/tmp", camera="bar", stop=2, delay=0)
        self.assertListEqual(dl_log.output, expected_log)

    @mock.patch("blinkpy.blinkpy.api.request_network_update")
    @mock.patch("blinkpy.auth.Auth.query")
    async def test_refresh(self, mock_req, mock_update):
        """Test ability to refresh system."""
        mock_update.return_value = {"network": {"sync_module_error": False}}
        mock_req.return_value.json = mock.AsyncMock(return_value={})
        self.blink.last_refresh = 0
        self.blink.available = True
        self.blink.auth.account_id = 1234
        self.blink.sync["foo"] = MockSyncModule(self.blink, "foo", 1, [])
        self.blink.cameras = {"bar": MockCamera(self.blink.sync)}
        self.blink.sync["foo"].cameras = self.blink.cameras
        self.assertTrue(await self.blink.refresh())

    @mock.patch("blinkpy.blinkpy.api.request_notification_flags")
    async def test_get_status(self, mock_req):
        """Test get of notification flags."""
        mock_req.return_value = {"notifications": {"foo": True}}
        self.assertDictEqual(await self.blink.get_status(), {"foo": True})

    @mock.patch("blinkpy.blinkpy.api.request_notification_flags")
    async def test_get_status_malformed(self, mock_req):
        """Test get of notification flags with malformed response."""
        mock_req.return_value = {"nobueno": {"foo": False}}
        self.assertDictEqual(await self.blink.get_status(), {"nobueno": {"foo": False}})

    @mock.patch("blinkpy.blinkpy.api.request_set_notification_flag")
    async def test_set_status(self, mock_req):
        """Test set of notification flags."""
        mock_req.return_value = True
        self.assertTrue(await self.blink.set_status())



"""Defines a sync module for Blink."""

import logging
import string
import datetime
import traceback
import asyncio
import aiofiles
from sortedcontainers import SortedSet
from requests.structures import CaseInsensitiveDict
from blinkpy import api
from blinkpy.camera import BlinkCamera, BlinkCameraMini, BlinkDoorbell
from blinkpy.helpers.util import (
    time_to_seconds,
    backoff_seconds,
    to_alphanumeric,
    json_dumps,
)
from blinkpy.helpers.constants import ONLINE

_LOGGER = logging.getLogger(__name__)


class BlinkSyncModule:
    """Class to initialize sync module."""

    def __init__(self, blink, network_name, network_id, camera_list):
        """
        Initialize Blink sync module.

        :param blink: Blink class instantiation
        """
        self.blink = blink
        self.network_id = network_id
        self.region_id = blink.auth.region_id
        self.name = network_name
        self.serial = None
        self._version = None
        self.status = "offline"
        self.sync_id = None
        self.host = None
        self.summary = None
        self.network_info = None
        self.events = []
        self.cameras = CaseInsensitiveDict({})
        self.motion_interval = blink.motion_interval
        self.motion = {}
        # A dictionary where keys are the camera names, and
        # values are lists of recent clips.
        self.last_records = {}
        self.camera_list = camera_list
        self.available = False
        # type_key_map is only for the mini's and the doorbells.
        # Outdoor cameras have their own URL API which must be queried.
        self.type_key_map = {
            "mini": "owls",
            "doorbell": "doorbells",
        }
        self._names_table = {}
        self._local_storage = {
            "enabled": False,
            "compatible": False,
            "status": False,
            "last_manifest_id": None,
            "manifest": SortedSet(),
            "manifest_stale": True,
            "last_manifest_read": datetime.datetime(1970, 1, 1, 0, 0, 0).isoformat(),
        }

    @property
    def attributes(self):
        """Return sync attributes."""
        attr = {
            "name": self.name,
            "id": self.sync_id,
            "network_id": self.network_id,
            "serial": self.serial,
            "version": self._version,
            "status": self.status,
            "region_id": self.region_id,
            "local_storage": self.local_storage,
        }
        return attr

    @property
    def urls(self):
        """Return device urls."""
        return self.blink.urls

    @property
    def online(self):
        """Return boolean system online status."""
        try:
            return ONLINE[self.status]
        except KeyError:
            _LOGGER.error("Unknown sync module status %s", self.status)
            self.available = False
            return False

    @property
    def version(self):
        """Return the Syncmodule Firmware version."""
        return self._version

    @property
    def arm(self):
        """Return status of sync module: armed/disarmed."""
        try:
            return self.network_info["network"]["armed"]
        except (KeyError, TypeError):
            self.available = False
            return None

    @property
    def local_storage(self):
        """Indicate if local storage is activated or not (True/False)."""
        return self._local_storage["status"]

    @property
    def local_storage_manifest_ready(self):
        """Indicate if the manifest is up-to-date."""
        return not self._local_storage["manifest_stale"]

    async def async_arm(self, value):
        """Arm or disarm camera."""
        if value:
            return await api.request_system_arm(self.blink, self.network_id)
        return await api.request_system_disarm(self.blink, self.network_id)

    async def start(self):
        """Initialize the system."""
        _LOGGER.debug("Initializing the sync module")
        response = await self.sync_initialize()
        if not response:
            return False

        try:
            self.sync_id = self.summary["id"]
            self.serial = self.summary["serial"]
            self.status = self.summary["status"]
        except KeyError:
            _LOGGER.error("Could not extract some sync module info: %s", response)

        is_ok = await self.get_network_info()

        if not is_ok or not await self.update_cameras():
            self.available = False
            return False
        self.available = True
        return True

    async def sync_initialize(self):
        """Initialize a sync module."""
        # Doesn't include local store info for some reason.
        response = await api.request_syncmodule(self.blink, self.network_id)
        try:
            self.summary = response["syncmodule"]
            self.network_id = self.summary["network_id"]
            await self._init_local_storage(self.summary["id"])
        except (TypeError, KeyError):
            _LOGGER.error(
                "Could not retrieve sync module information with response: %s", response
            )
            return False
        self._version = self.summary.get("fw_version")
        return response

    async def _init_local_storage(self, sync_id):
        """Initialize local storage from homescreen dictionary."""
        home_screen = self.blink.homescreen
        sync_module = None
        try:
            sync_modules = home_screen["sync_modules"]
            for mod in sync_modules:
                if mod["id"] == sync_id:
                    self._local_storage["enabled"] = mod["local_storage_enabled"]
                    self._local_storage["compatible"] = mod["local_storage_compatible"]
                    self._local_storage["status"] = (
                        mod["local_storage_status"] == "active"
                    )
                    self._local_storage["last_manifest_read"] = (
                        datetime.datetime.utcnow() - datetime.timedelta(seconds=10)
                    ).isoformat()
                    sync_module = mod
        except (TypeError, KeyError):
            _LOGGER.error(
                "Could not retrieve sync module information from home screen: %s",
                home_screen,
            )
            return False
        return sync_module

    async def update_cameras(self, camera_type=BlinkCamera):
        """Update cameras from server."""
        type_map = {
            "mini": BlinkCameraMini,
            "doorbell": BlinkDoorbell,
            "default": BlinkCamera,
        }
        try:
            _LOGGER.debug("Updating cameras")
            for camera_config in self.camera_list:
                _LOGGER.debug("Updating camera_config %s", json_dumps(camera_config))
                if "name" not in camera_config:
                    break
                blink_camera_type = camera_config.get("type", "")
                name = camera_config["name"]
                self.motion[name] = False
                unique_info = self.get_unique_info(name)
                if blink_camera_type in type_map:
                    camera_type = type_map[blink_camera_type]
                self.cameras[name] = camera_type(self)
                camera_info = await self.get_camera_info(
                    camera_config["id"], unique_info=unique_info
                )
                self._names_table[to_alphanumeric(name)] = name
                await self.cameras[name].update(
                    camera_info, force_cache=True, force=True
                )
        except KeyError:
            _LOGGER.error("Could not create camera instances for %s", self.name)
            return False
        return True

    def get_unique_info(self, name):
        """Extract unique information for Minis and Doorbells."""
        try:
            for type_key in self.type_key_map.values():
                for device in self.blink.homescreen[type_key]:
                    _LOGGER.debug("checking device %s", device)
                    if device["name"] == name:
                        _LOGGER.debug("Found unique_info %s", device)
                        return device
        except (TypeError, KeyError):
            pass
        return None

    async def get_events(self, **kwargs):
        """Retrieve events from server."""
        force = kwargs.pop("force", False)
        response = await api.request_sync_events(
            self.blink, self.network_id, force=force
        )
        try:
            return response["event"]
        except (TypeError, KeyError):
            _LOGGER.error("Could not extract events: %s", response)
            return False

    async def get_camera_info(self, camera_id, **kwargs):
        """Retrieve camera information."""
        unique = kwargs.get("unique_info", None)
        if unique is not None:
            return unique
        response = await api.request_camera_info(self.blink, self.network_id, camera_id)
        try:
            return response["camera"][0]
        except (TypeError, KeyError):
            _LOGGER.error(
                "Could not extract camera info for %s: %s", camera_id, response
            )
            return {}

    async def get_network_info(self):
        """Retrieve network status."""
        self.network_info = await api.request_network_update(
            self.blink, self.network_id
        )
        try:
            if self.network_info["network"]["sync_module_error"]:
                raise KeyError
        except (TypeError, KeyError):
            self.available = False
            return False
        return True

    async def refresh(self, force_cache=False):
        """Get all blink cameras and pulls their most recent status."""
        if not await self.get_network_info():
            return
        await self.update_local_storage_manifest()
        await self.check_new_videos()
        for camera_name in self.cameras:
            camera_id = self.cameras[camera_name].camera_id
            camera_info = await self.get_camera_info(
                camera_id,
                unique_info=self.get_unique_info(camera_name),
            )
            await self.cameras[camera_name].update(camera_info, force_cache=force_cache)
        self.available = True

    async def check_new_videos(self):
        """Check if new videos since last refresh."""
        _LOGGER.debug("Checking for new videos")
        try:
            interval = self.blink.last_refresh - self.motion_interval * 60
            last_refresh = datetime.datetime.fromtimestamp(self.blink.last_refresh)
            _LOGGER.debug("last_refresh = %s", last_refresh)
            _LOGGER.debug("interval = %s", interval)
        except TypeError:
            # This is the first start, so refresh hasn't happened yet.
            # No need to check for motion.
            ex = traceback.format_exc()
            _LOGGER.error(
                "Error calculating interval (last_refresh = %s): %s",
                self.blink.last_refresh,
                ex,
            )
            trace = "".join(traceback.format_stack())
            _LOGGER.debug("\n%s", trace)
            _LOGGER.info("No new videos since last refresh.")
            return False

        resp = await api.request_videos(self.blink, time=interval, page=1)

        last_record = {}
        for camera in self.cameras:
            # Initialize the list if doesn't exist yet.
            if camera not in self.last_records:
                self.last_records[camera] = []
            # Hang on to the last record if there is one.
            if len(self.last_records[camera]) > 0:
                last_record[camera] = self.last_records[camera][-1]
            # Reset in preparation for processing new entries.
            self.last_records[camera] = []
            self.motion[camera] = False

        try:
            info = resp["media"]
        except (KeyError, TypeError):
            _LOGGER.warning("Could not check for motion. Response: %s", resp)
            return False

        for entry in info:
            try:
                name = entry["device_name"]
                clip_url = entry["media"]
                timestamp = entry["created_at"]
                if self.check_new_video_time(timestamp):
                    self.motion[name] = True and self.arm
                    record = {"clip": clip_url, "time": timestamp}
                    self.last_records[name].append(record)
            except KeyError:
                last_refresh = datetime.datetime.fromtimestamp(self.blink.last_refresh)
                _LOGGER.debug(
                    "No new videos for %s since last refresh at %s.",
                    entry,
                    last_refresh,
                )

        # Process local storage if active and if the manifest is ready.
        last_manifest_read = datetime.datetime.fromisoformat(
            self._local_storage["last_manifest_read"]
        )
        _LOGGER.debug("last_manifest_read = %s", last_manifest_read)
        _LOGGER.debug("Manifest ready? %s", self.local_storage_manifest_ready)
        if self.local_storage and self.local_storage_manifest_ready:
            _LOGGER.debug("Processing updated manifest")
            manifest = self._local_storage["manifest"]
            last_manifest_id = self._local_storage["last_manifest_id"]
            last_manifest_read = self._local_storage["last_manifest_read"]
            last_read_local = (
                datetime.datetime.fromisoformat(last_manifest_read)
                .replace(tzinfo=datetime.timezone.utc)
                .astimezone(tz=None)
            )
            last_clip_time = None
            num_new = 0
            for item in reversed(manifest):
                iso_timestamp = item.created_at.isoformat()

                _LOGGER.debug(
                    "Checking '%s': clip_time = %s, manifest_read = %s",
                    item.name,
                    iso_timestamp,
                    last_manifest_read,
                )
                # Exit the loop once there are no new videos in the list.
                if not self.check_new_video_time(iso_timestamp, last_manifest_read):
                    _LOGGER.info(
                        "No new local storage videos since last manifest "
                        "read at %s.",
                        last_read_local,
                    )
                    break
                _LOGGER.debug("Found new item in local storage manifest: %s", item)
                name = item.name
                clip_url = item.url(last_manifest_id)
                await item.prepare_download(self.blink)
                self.motion[name] = True
                record = {"clip": clip_url, "time": iso_timestamp}
                self.last_records[name].append(record)
                last_clip_time = item.created_at
                num_new += 1

            # The manifest became ready, and we read recent clips from it.
            if num_new > 0:
                last_manifest_read = (
                    datetime.datetime.utcnow() - datetime.timedelta(seconds=10)
                ).isoformat()
                self._local_storage["last_manifest_read"] = last_manifest_read
                _LOGGER.debug("Updated last_manifest_read to %s", last_manifest_read)
                _LOGGER.debug("Last clip time was %s", last_clip_time)
        # We want to keep the last record when no new motion was detected.
        for camera in self.cameras:
            # Check if there are no new records, indicating motion.
            if len(self.last_records[camera]) == 0:
                # If no new records, check if we had a previous last record.
                if camera in last_record:
                    # Put the last record back into the empty list.
                    self.last_records[camera].append(last_record[camera])

        return True

    def check_new_video_time(self, timestamp, reference=None):
        """Check if video has timestamp since last refresh.

        :param timestamp ISO-formatted timestamp string
        :param reference ISO-formatted reference timestamp string
        """
        if not reference:
            return time_to_seconds(timestamp) > self.blink.last_refresh
        return time_to_seconds(timestamp) > time_to_seconds(reference)

    async def update_local_storage_manifest(self):
        """Update local storage manifest, which lists all stored clips."""
        if not self.local_storage:
            self._local_storage["manifest_stale"] = True
            return None
        _LOGGER.debug("Updating local storage manifest")

        response = await self.poll_local_storage_manifest()
        try:
            manifest_request_id = response["id"]
        except (TypeError, KeyError):
            _LOGGER.error(
                "Could not extract manifest request ID from response: %s", response
            )
            self._local_storage["manifest_stale"] = True
            return None

        response = await self.poll_local_storage_manifest(manifest_request_id)
        try:
            manifest_id = response["manifest_id"]
        except (TypeError, KeyError):
            _LOGGER.error("Could not extract manifest ID from response: %s", response)
            self._local_storage["manifest_stale"] = True
            return None

        self._local_storage["last_manifest_id"] = manifest_id
        template = string.Template(api.local_storage_clip_url_template()).substitute(
            account_id=self.blink.account_id,
            network_id=self.network_id,
            sync_id=self.sync_id,
            manifest_id="$manifest_id",
            clip_id="$clip_id",
        )
        num_stored = len(self._local_storage["manifest"])
        try:
            for item in response["clips"]:
                alphanumeric_name = item["camera_name"]
                if alphanumeric_name in self._names_table:
                    camera_name = self._names_table[alphanumeric_name]
                    self._local_storage["manifest"].add(
                        LocalStorageMediaItem(
                            item["id"],
                            camera_name,
                            item["created_at"],
                            item["size"],
                            manifest_id,
                            template,
                        )
                    )
            num_added = len(self._local_storage["manifest"]) - num_stored
            if num_added > 0:
                _LOGGER.info(
                    "Found %s new clip(s) in local storage manifest id = %s",
                    num_added,
                    manifest_id,
                )
        except (TypeError, KeyError):
            ex = traceback.format_exc()
            _LOGGER.error("Could not extract clips list from response: %s", ex)
            trace = "".join(traceback.format_stack())
            _LOGGER.debug("\n%s", trace)
            self._local_storage["manifest_stale"] = True
            return None

        self._local_storage["manifest_stale"] = False
        return True

    async def poll_local_storage_manifest(
        self, manifest_request_id=None, max_retries=4
    ):
        """Poll for local storage manifest."""
        # The sync module may be busy processing another request
        # (like saving a new clip).
        # Poll the endpoint until it is ready, backing off each retry.
        response = None
        for retry in range(max_retries):
            # Request building the manifest.
            if not manifest_request_id:
                response = await api.request_local_storage_manifest(
                    self.blink, self.network_id, self.sync_id
                )
                if response and "id" in response:
                    break
            # Get the manifest.
            else:
                response = await api.get_local_storage_manifest(
                    self.blink, self.network_id, self.sync_id, manifest_request_id
                )
                if response and "clips" in response:
                    break
            seconds = backoff_seconds(retry=retry, default_time=3)
            _LOGGER.debug("[retry=%d] Retrying in %d seconds", retry + 1, seconds)
            await asyncio.sleep(seconds)
        return response


class BlinkOwl(BlinkSyncModule):
    """Representation of a sync-less device."""

    def __init__(self, blink, name, network_id, response):
        """Initialize a sync-less object."""
        cameras = [{"name": name, "id": response["id"]}]
        super().__init__(blink, name, network_id, cameras)
        self.sync_id = response["id"]
        self.serial = response["serial"]
        self.status = response["enabled"]
        if not self.serial:
            self.serial = f"{network_id}-{self.sync_id}"

    async def sync_initialize(self):
        """Initialize a sync-less module."""
        self.summary = {
            "id": self.sync_id,
            "name": self.name,
            "serial": self.serial,
            "status": self.status,
            "onboarded": True,
            "account_id": self.blink.account_id,
            "network_id": self.network_id,
        }
        return self.summary

    async def update_cameras(self, camera_type=BlinkCameraMini):
        """Update sync-less cameras."""
        return await super().update_cameras(camera_type=BlinkCameraMini)

    async def get_camera_info(self, camera_id, **kwargs):
        """Retrieve camera information."""
        try:
            for owl in self.blink.homescreen["owls"]:
                if owl["name"] == self.name:
                    self.status = owl["enabled"]
                    return owl
        except (TypeError, KeyError):
            pass
        return None

    async def get_network_info(self):
        """Get network info for sync-less module."""
        return True

    @property
    def network_info(self):
        """Format owl response to resemble sync module."""
        return {
            "network": {
                "id": self.network_id,
                "name": self.name,
                "armed": self.status,
                "sync_module_error": False,
                "account_id": self.blink.account_id,
            }
        }

    @network_info.setter
    def network_info(self, value):
        """Set network_info property."""


class BlinkLotus(BlinkSyncModule):
    """Representation of a sync-less device."""

    def __init__(self, blink, name, network_id, response):
        """Initialize a sync-less object."""
        cameras = [{"name": name, "id": response["id"]}]
        super().__init__(blink, name, network_id, cameras)
        self.sync_id = response["id"]
        self.serial = response["serial"]
        self.status = response["enabled"]
        if not self.serial:
            self.serial = f"{network_id}-{self.sync_id}"

    async def sync_initialize(self):
        """Initialize a sync-less module."""
        self.summary = {
            "id": self.sync_id,
            "name": self.name,
            "serial": self.serial,
            "status": self.status,
            "onboarded": True,
            "account_id": self.blink.account_id,
            "network_id": self.network_id,
        }
        return self.summary

    async def update_cameras(self, camera_type=BlinkDoorbell):
        """Update sync-less cameras."""
        return await super().update_cameras(camera_type=BlinkDoorbell)

    async def get_camera_info(self, camera_id, **kwargs):
        """Retrieve camera information."""
        try:
            for doorbell in self.blink.homescreen["doorbells"]:
                if doorbell["name"] == self.name:
                    self.status = doorbell["enabled"]
                    return doorbell
        except (TypeError, KeyError):
            pass
        return None

    async def get_network_info(self):
        """Get network info for sync-less module."""
        return True

    @property
    def network_info(self):
        """Format lotus response to resemble sync module."""
        return {
            "network": {
                "id": self.network_id,
                "name": self.name,
                "armed": self.status,
                "sync_module_error": False,
                "account_id": self.blink.account_id,
            }
        }

    @network_info.setter
    def network_info(self, value):
        """Set network_info property."""


class LocalStorageMediaItem:
    """Metadata of media item in the local storage manifest."""

    def __init__(
        self, item_id, camera_name, created_at, size, manifest_id, url_template
    ):
        """Initialize media item.

        :param item_id: ID of the manifest item.
        :param camera_name: Name of camera that took the video.
        :param created_at: ISO-formatted time stamp for creation time.
        :param size: Size of the video file.
        """
        self._id = int(item_id)
        self._camera_name = camera_name
        self._created_at = datetime.datetime.fromisoformat(created_at)
        self._size = size
        self._url_template = url_template
        self._manifest_id = manifest_id

    def _build_url(self, manifest_id, clip_id):
        return string.Template(self._url_template).substitute(
            manifest_id=manifest_id, clip_id=clip_id
        )

    @property
    def id(self):
        """Return media item ID."""
        return self._id

    @property
    def name(self):
        """Return name of camera that captured this media item."""
        return self._camera_name

    @property
    def created_at(self):
        """Return the ISO-formatted creation time stamp of this media item."""
        return self._created_at

    @property
    def size(self):
        """Return the reported size of this media item."""
        return self._size

    def url(self, manifest_id=None):
        """Build the URL.

        Builds the url new each time since the media item is cached,
        and the manifest is possibly rebuilt each refresh.

        :param manifest_id: ID of new manifest (if it changed)
        :return: URL for clip retrieval
        """
        if manifest_id:
            self._manifest_id = manifest_id
        return self._build_url(self._manifest_id, self._id)

    async def prepare_download(self, blink, max_retries=4):
        """Initiate upload of media item from the sync module to Blink cloud servers."""
        if max_retries == 0:
            return None
        url = blink.urls.base_url + self.url()
        response = await api.http_post(blink, url)
        await api.wait_for_command(blink, response)
        return response

    async def delete_video(self, blink, max_retries=4) -> bool:
        """Delete video from sync module."""
        delete_url = blink.urls.base_url + self.url()
        delete_url = delete_url.replace("request", "delete")

        for retry in range(max_retries):
            delete = await api.http_post(
                blink, delete_url, json=False
            )  # Delete the video
            if delete.status == 200:
                return True
            seconds = backoff_seconds(retry=retry, default_time=3)
            _LOGGER.debug("[retry=%d] Retrying in %d seconds", retry + 1, seconds)
            await asyncio.sleep(seconds)
        return False

    async def download_video(self, blink, file_name, max_retries=4) -> bool:
        """Download a previously prepared video from sync module."""
        for retry in range(max_retries):
            url = blink.urls.base_url + self.url()
            video = await api.http_get(blink, url, json=False)
            if video.status == 200:
                async with aiofiles.open(file_name, "wb") as vidfile:
                    await vidfile.write(await video.read())  # download the video
                    return True
            seconds = backoff_seconds(retry=retry, default_time=3)
            _LOGGER.debug(
                "[retry=%d] Retrying in %d seconds: %s", retry + 1, seconds, url
            )
            await asyncio.sleep(seconds)
        return False

    async def download_video_delete(self, blink, file_name, max_retries=4) -> bool:
        """Delete local videos.

        Initiate upload of media item from the sync module to
        Blink cloud servers then download to local filesystem and delete from sync.
        """
        if await self.prepare_download(blink):
            if await self.download_video(blink, file_name):
                if await self.delete_video(blink):
                    return True
        return False

    def __repr__(self):
        """Create string representation."""
        return (
            f"LocalStorageMediaItem(id={self._id}, camera_name={self._camera_name}, "
            f"created_at={self._created_at}"
            + f", size={self._size}, manifest_id={self._manifest_id}, "
            f"url_template={self._url_template})"
        )

    def __str__(self):
        """Create string representation."""
        return self.__repr__()

    def cmp_key(self):
        """Return key to use for comparison."""
        return self._created_at

    def __eq__(self, other):
        """Check equality."""
        return self.cmp_key() == other.cmp_key()

    def __lt__(self, other):
        """Check less than."""
        return self.cmp_key() < other.cmp_key()

    def __hash__(self):
        """Return unique hash value."""
        return self._id
